/**
 * @file test_pair_ZZZ.c
 * @author Alessandro Budroni
 * @brief Test function for pairing
 *
 * LICENSE
 *
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pair_ZZZ.h"
#include "randapi.h"

#define N_ITER 16

int main()
{

    int i;
    BIG_XXX s,r,x,y;
    ECP_ZZZ G,P1,P2,P3;
    ECP2_ZZZ W,Q1,Q2,Q3;
    FP12_YYY g1,g2;
    FP2_YYY wx,wy;

    char seed[100] = {0};
    octet SEED = {0,sizeof(seed),seed};
    csprng RNG;

    /* unrandom seed value! */
    SEED.len=100;
    for (i=0; i<100; i++) SEED.val[i]=i+1;

    /* initialise random number generator */
    CREATE_CSPRNG(&RNG,&SEED);

    // Set Generator of G1
    BIG_XXX_rcopy(x,CURVE_Gx_ZZZ);
    BIG_XXX_rcopy(y,CURVE_Gy_ZZZ);
    ECP_ZZZ_set(&G,x,y);
    if (G.inf)
    {
        printf("Failed to set - point not on curve\n");
        return 1;
    }

    // Set Generator of G2
    FP_YYY_rcopy(&(wx.a),CURVE_Pxa_ZZZ);
    FP_YYY_rcopy(&(wx.b),CURVE_Pxb_ZZZ);
    FP_YYY_rcopy(&(wy.a),CURVE_Pya_ZZZ);
    FP_YYY_rcopy(&(wy.b),CURVE_Pyb_ZZZ);
    ECP2_ZZZ_set(&W,&wx,&wy);
    if (W.inf)
    {
        printf("Failed to set - point not on curve\n");
        return 1;
    }

    // Check that scalar multiplication by curve order gives the infinte point
    BIG_XXX_rcopy(r,CURVE_Order_ZZZ);
    ECP_ZZZ_copy(&P1,&G);
    PAIR_ZZZ_G1mul(&P1,r);
    if (!ECP_ZZZ_isinf(&P1))
    {
        printf("FAILURE - rP!=O\n");
        return 1;
    }

    BIG_XXX_randomnum(s,r,&RNG);
    ECP2_ZZZ_copy(&Q1,&W);
    ECP2_ZZZ_mul(&Q1,r);

    if (!ECP2_ZZZ_isinf(&Q1))
    {
        printf("FAILURE - rQ!=O\n");
        return 0;
    }

    for (i = 0; i < N_ITER; ++i)
    {
        // Pick a random point in G1
        BIG_XXX_randomnum(x,r,&RNG);
        ECP_ZZZ_copy(&P1,&G);
        ECP_ZZZ_mul(&P1,x);
        ECP_ZZZ_copy(&P2,&P1);

        // Pick a random point in G2
        BIG_XXX_randomnum(y,r,&RNG);
        ECP2_ZZZ_copy(&Q1,&W);
        ECP2_ZZZ_mul(&Q1,y);
        ECP2_ZZZ_copy(&Q2,&Q1);

        // Test that e(sQ,P) = e(Q,sP) = e(Q,P)^s, s random
        BIG_XXX_randomnum(s,r,&RNG);
        PAIR_ZZZ_G1mul(&P2,s);
        ECP2_ZZZ_mul(&Q2,s);
        PAIR_ZZZ_ate(&g1,&Q1,&P2);
        PAIR_ZZZ_fexp(&g1);
        PAIR_ZZZ_ate(&g2,&Q2,&P1);
        PAIR_ZZZ_fexp(&g2);
        if (!FP12_YYY_equals(&g1,&g2))
        {
            printf("FAILURE - e(sQ,P)!=e(Q,sP)\n");
            return 1;
        }
        PAIR_ZZZ_ate(&g2,&Q1,&P1);
        PAIR_ZZZ_fexp(&g2);
        PAIR_ZZZ_GTpow(&g2,s);
        if (!FP12_YYY_equals(&g1,&g2))
        {
            printf("FAILURE - e(sQ,P)!=e(Q,P)^s\n");
            return 1;
        }

        // Test that e(Q,P1+P2) = e(Q,P1).e(Q,P2)
        PAIR_ZZZ_ate(&g1,&Q1,&P1);
        PAIR_ZZZ_fexp(&g1);
        PAIR_ZZZ_ate(&g2,&Q1,&P2);
        PAIR_ZZZ_fexp(&g2);
        FP12_YYY_mul(&g1,&g2);
        ECP_ZZZ_copy(&P3,&P1);
        ECP_ZZZ_add(&P3,&P2);
        ECP_ZZZ_affine(&P3);
        PAIR_ZZZ_ate(&g2,&Q1,&P3);
        PAIR_ZZZ_fexp(&g2);
        if (!FP12_YYY_equals(&g1,&g2))
        {
            printf("FAILURE - e(Q,P1+P2)!=e(Q,P1).e(Q,P2)\n");
            return 1;
        }

        // Test double_ate function on G1
        PAIR_ZZZ_double_ate(&g1,&Q1,&P1,&Q1,&P2);
        PAIR_ZZZ_fexp(&g1);
        if (!FP12_YYY_equals(&g1,&g2))
        {
            printf("FAILURE - double ate failed\n");
            return 1;
        }

        // Test that e(Q1+Q2,P1) = e(Q1,P1).e(Q2,P1)
        PAIR_ZZZ_ate(&g1,&Q1,&P1);
        PAIR_ZZZ_fexp(&g1);
        PAIR_ZZZ_ate(&g2,&Q2,&P1);
        PAIR_ZZZ_fexp(&g2);
        FP12_YYY_mul(&g1,&g2);
        ECP2_ZZZ_copy(&Q3,&Q1);
        ECP2_ZZZ_add(&Q3,&Q2);
        ECP2_ZZZ_affine(&Q3);
        PAIR_ZZZ_ate(&g2,&Q3,&P1);
        PAIR_ZZZ_fexp(&g2);
        if (!FP12_YYY_equals(&g1,&g2))
        {
            printf("FAILURE - e(Q1+Q2,P1)!=e(Q1,P1).e(Q2,P1)\n");
            return 1;
        }

        // Test double_ate function on G2
        PAIR_ZZZ_double_ate(&g1,&Q1,&P1,&Q2,&P1);
        PAIR_ZZZ_fexp(&g1);
        if (!FP12_YYY_equals(&g1,&g2))
        {
            printf("FAILURE - double ate failed\n");
            return 1;
        }

    }

    KILL_CSPRNG(&RNG);
    printf("SUCCESS\n");
    return 0;
}
